/*-
 * Copyright (c) 2008 FUKAUMI Naoki.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <glib.h>
#include <math.h>
#include <stdio.h>
#include "usbops.hpp"
#include "Hdus.hpp"

#include "decrypt.c"

Hdus::Hdus()
    : AbstractFriio(), saddr( 0 ), buffer( NULL )
{
	supportBands.push_back(BAND_UHF);
	targetType         = TUNER_HDUS;
	asyncBufSize       = HDUS_ASYNCBUFFSIZE;
	requestReserveNum  = HDUS_REQUEST_RESERVE_NUM;
	requestPollingWait = HDUS_REQUEST_POLLING_WAIT;
}


Hdus::~Hdus()
{
    if( initialized ) close();
    if( buffer ) free( buffer );
    buffer = NULL;
}


/**
 * 周波数帯域/チャンネルを設定する。
 * @param newBand 周波数帯域 UHF/BS/CS
 * @param newChannel チャンネル
 */
void
Hdus::setChannel(BandType newBand, int newChannel)
{
    if( ! saddr ) return;

    int ugen = tunerFd;
    uint8_t data[0x200];
    int chan = newChannel, freq, val;
    int xpid_min = 0x1fff, xpid_max = 0x1fff;

    if (BAND_UHF != newBand || newChannel < 13 || newChannel > 62) {
        throw traceable_error("unknown channel."); // TODO: 適当なエラークラス
    }
    freq = 473 + (chan - 13) * 6;
    if (log) *log << "freq = " << freq << std::endl;


    /* Set frequency */
    val = (freq + 57) * 7;
    usleep (50 * 1000);
    send_ctl(data, ugen, 0x0d, 0xfe00, (val & 0xff00) | 0xc0, 0x0004);
    send_ctl(data, ugen, 0x0d, ((val & 0xff) << 8) | 0x03,
	     (freq < 420) ? 0x8291 : 0x8491, 0x0004);
    send_ctl(data, ugen, 0x0d, 0xe306, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x0e, 0x0020, 0x0000, 0x0008);
    send_ctl(data, ugen, 0x03, 0x0120, 0x0040, 0x0002);

    send_ctl(data, ugen, 0x09, 0x0100, 0x0000, 0x0001);

    /* pid filter configuration */
    send_ctl(data, ugen, 0x04, 0x0040, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x05, ((data[1] | 0x0f) << 8) | 0x40, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x05, (xpid_min & 0xff00) | 0x41,
	     (xpid_min & 0xff), 0x0003);
    send_ctl(data, ugen, 0x05, (xpid_max & 0xff00) | 0x43,
	     (xpid_max & 0xff), 0x0003);

    /* ASIE5606 decrypt key configuration */
    int i;
    for (i = 0; i < 0x10; i++)
      send_ctl(data, ugen, 0x03, ((i + 0x10) << 8) | saddr, key[i], 0x0002);

    /* kick ASV5211 bulk transfer */
    send_ctl(data, ugen, 0x06, 0x0000, 0x0000, 0x0001);
}

/**
 * 信号レベルを取得する。
 * @return float 信号レベル
 */
const float
Hdus::getSignalLevel(void)
{
    /* 取得方法が分からないので、とりあえず THRESHOLD より大きい値を返しておく */
    return  SIGNALLEVEL_RETRY_THRESHOLD + 8;
}


/**
 * データを取得する
 * @param bufp データへのポインタを取得するポインタ(出力)
 *             タイムアウト時NULL
 * @param timeoutMsec データを待つ時間(msec)
 *                    0の場合タイムアウト無し
 * @return int 取得されたデータ長
 */
int
Hdus::getStream(const uint8_t** bufp, int timeoutMsec)
{
    if( ! buffer  ){
        *bufp = NULL;
        return 0;
    }

    const size_t tssize = 188;

    const int readsize = AbstractFriio::getStream( bufp, timeoutMsec );
    if( *bufp == NULL ) return readsize;
    if( ! readsize ) return readsize;

    /* XOR 処理 */

//    fprintf( stderr, "buffersize = %d, remainsize = %d", buffersize, buffer_remainsize );

    // 前回の処理の残りをコピー
    if( buffer_remainsize ){
        memmove( buffer, buffer + ( buffersize - buffer_remainsize ), buffer_remainsize );
        buffersize = buffer_remainsize;
        buffer_remainsize = 0;
    }
    else buffersize = 0;

    memcpy( buffer + buffersize, *bufp, readsize );
    *bufp = buffer;
    buffersize += readsize;

//    fprintf( stderr, " -> buffersize = %d, remainsize = %d", buffersize, buffer_remainsize );

    size_t pos = 0;
    while( 1 ){

        // 同期
        size_t pos_sync = pos;
        while( pos < buffersize && buffer[ pos ] != 0x47 ) ++pos;
        if( pos_sync != pos ){
            if (log) *log << "sync " << pos - pos_sync << " bytes" << std::endl;
        }
        if( buffersize - pos < tssize ){
            buffer_remainsize = buffersize - pos;
            break;
        }

        // XOR
        uint32_t *buf = (uint32_t*)( buffer + pos );
	static unsigned char tmp[188];
	memcpy(tmp, buf, 188);
	// src, dest, packet_num
        packetDecrypt(tmp, (uchar *)buf, 1);
        // for( int i = 1; i < 47; i++)  buf[i] ^= xxor[i];

        pos += tssize;
    }

//    fprintf( stderr, "-> pos = %d\n", pos );

    return pos;
}


/**
 * デバイスファイルがFriioの物かを確認する。
 * @param devfile デバイスファイル
 * @return Friioである場合true;
 */
bool
Hdus::is_friio(const std::string &devfile)
{
	usb_device_descriptor usb_desc;
	try {
		usb_getdesc(devfile.c_str(), &usb_desc);
	} catch (usb_error &e) {
		if (log) *log << "usb_getdesc: " << e.what() << std::endl;
		return false;
	}
	
	if ( TARGET_ID_VENDOR_HDUS == usb_desc.idVendor && TARGET_ID_PRODUCT_HDUS == usb_desc.idProduct) {
		return true;
	} else {
		return false;
	}
}


/**
 * 固定処理A(デバイス判定)
 * @param fd 対象ファイルディスクリプタ
 * @return TunerType チューナの種別
 */
TunerType
Hdus::UsbProcFixInitA(int fd)
{
    return TUNER_HDUS;
}


/**
 * XOR テーブル読み込み
*/
void
Hdus::LoadXorTable()
{
    size_t syssize;
    uint32_t *xp;
    uint8_t *hp;
    size_t offset;
    int i;
    FILE* xfd = NULL;

    if (log) *log << "reading SKNET_HDTV_BDA.sys...";

    if ( ! ( xfd = fopen("SKNET_HDTV_BDA.sys", "rb" ) ) ) throw io_error( "could not open SKNET_HDTV_BDA.sys" );

    syssize = 0x12da8 + 184;
    hp = (uint8_t*)malloc( syssize );
    if( fread( hp, 1, syssize, xfd ) != syssize ){
        fclose(xfd);
        free( hp );
        throw io_error( "could not read SKNET_HDTV_BDA.sys" );
    }

    if (hp[0x2c8] == 0x8d && hp[0x2c9] == 0x5f)
        offset = 0xf628; /* CD 1.0 */
    else if (hp[0x2c0] == 0x31 && hp[0x2c1] == 0xe6)
        offset = 0x12da8; /* CD 1.1 */
    else {
        fclose(xfd);
        free( hp );
        throw io_error( "unknown file" );
    }

    xp = (uint32_t*)(hp + offset);
    memset(xxor, 0, sizeof(xxor));
    for (i = 1; i < 47; i++) xxor[i] = xp[i - 1];

    fclose(xfd);
    free( hp );

    if (log) *log << "done. offset = " << offset << std::endl;
}


/**
 * 初期化処理 : 処理開始
 * 便宜上Beginにしているが、本来どんなことをしているかは不明
 * @param fd 対象ファイルディスクリプタ
 */
void
Hdus::UsbProcBegin(int fd)
{
    if( ! buffer ){

        LoadXorTable();
        buffer = (uint8_t*)malloc( HDUS_ASYNCBUFFSIZE * TSDATASIZE + 1024 );
    }

    buffersize = 0;
    buffer_remainsize = 0;

    /* HDUS を開く */

    if (log) *log << "opening HDUS...";

    int ugen = fd;
    uint8_t data[0x200], r5 = 0x91;
    saddr = 0x5a;

    send_ctl(data, ugen, 0x0c, 0x0000, 0x0000, 0x003a);
    send_ctl(data, ugen, 0x17, 0x0002, 0x0000, 0x0003);
    send_ctl(data, ugen, 0x08, 0xfb07, 0x0000, 0x0001);
    send_ctl(data, ugen, 0x10, 0x0200, 0x0000, 0x0001);

    /* Power off */
    usleep (50 * 1000);
    send_ctl(data, ugen, 0x08, 0x0100, 0x0000, 0x0001);
    usleep (100 * 1000);
    send_ctl(data, ugen, 0x08, 0xdd0c, 0x0000, 0x0001);
    usleep (100 * 1000);

    send_ctl(data, ugen, 0x17, 0x0002, 0x0000, 0x0003);

    /* Tuner on */
    send_ctl(data, ugen, 0x08, 0xd500, 0x0000, 0x0001);
    usleep (50 * 1000);
    send_ctl(data, ugen, 0x08, 0x0800, 0x0000, 0x0001);
    send_ctl(data, ugen, 0x08, 0x4040, 0x0000, 0x0001);
    usleep (5 * 1000);
    send_ctl(data, ugen, 0x08, 0x8080, 0x0000, 0x0001);
    usleep (50 * 1000);
    send_ctl(data, ugen, 0x08, 0x0404, 0x0000, 0x0001);
    usleep (50 * 1000);

    /* Tuner init */
    /* Tuner init */
    send_ctl(data, ugen, 0x02, 0x00a8, 0x0000, 0x0002);
    /* */
    send_ctl(data, ugen, 0x03, 0x0120, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x0220, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x0320, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x0420, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x0520, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x0620, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x0720, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x0820, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x0c20, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x1120, 0x0021, 0x0002);
    send_ctl(data, ugen, 0x03, 0x1220, 0x000b, 0x0002);
    send_ctl(data, ugen, 0x03, 0x1320, 0x002c, 0x0002);
    send_ctl(data, ugen, 0x03, 0x1420, 0x000a, 0x0002);
    send_ctl(data, ugen, 0x03, 0x1520, 0x0040, 0x0002);
    send_ctl(data, ugen, 0x03, 0x1620, 0x000a, 0x0002);
    send_ctl(data, ugen, 0x03, 0x1720, 0x0070, 0x0002);
    send_ctl(data, ugen, 0x03, 0x1820, 0x0031, 0x0002);
    send_ctl(data, ugen, 0x03, 0x1920, 0x0013, 0x0002);
    send_ctl(data, ugen, 0x03, 0x1a20, 0x0031, 0x0002);
    send_ctl(data, ugen, 0x03, 0x1b20, 0x0013, 0x0002);
    send_ctl(data, ugen, 0x03, 0x1c20, 0x002a, 0x0002);
    send_ctl(data, ugen, 0x03, 0x1d20, 0x00aa, 0x0002);
    send_ctl(data, ugen, 0x03, 0x1e20, 0x00a2, 0x0002);
    send_ctl(data, ugen, 0x03, 0x1f20, 0x0008, 0x0002);
    send_ctl(data, ugen, 0x03, 0x2020, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x2120, 0x00ff, 0x0002);
    send_ctl(data, ugen, 0x03, 0x2220, 0x0080, 0x0002);
    send_ctl(data, ugen, 0x03, 0x2320, 0x004c, 0x0002);
    send_ctl(data, ugen, 0x03, 0x2420, 0x004c, 0x0002);
    send_ctl(data, ugen, 0x03, 0x2520, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x2620, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x2720, 0x000c, 0x0002);
    send_ctl(data, ugen, 0x03, 0x2820, 0x0060, 0x0002);
    send_ctl(data, ugen, 0x03, 0x2920, 0x006b, 0x0002);
    send_ctl(data, ugen, 0x03, 0x2a20, 0x0040, 0x0002);
    send_ctl(data, ugen, 0x03, 0x2b20, 0x0040, 0x0002);
    send_ctl(data, ugen, 0x03, 0x2c20, 0x00ff, 0x0002);
    send_ctl(data, ugen, 0x03, 0x2d20, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x2e20, 0x00ff, 0x0002);
    send_ctl(data, ugen, 0x03, 0x2f20, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x3020, 0x0028, 0x0002);
    send_ctl(data, ugen, 0x03, 0x3120, 0x000f, 0x0002);
    send_ctl(data, ugen, 0x03, 0x3220, 0x00a0, 0x0002);
    send_ctl(data, ugen, 0x03, 0x3420, 0x003f, 0x0002);
    send_ctl(data, ugen, 0x03, 0x3820, 0x0001, 0x0002);
    send_ctl(data, ugen, 0x03, 0x3920, 0x0070, 0x0002);
    send_ctl(data, ugen, 0x03, 0x3a20, 0x002d, 0x0002);
    send_ctl(data, ugen, 0x03, 0x3b20, 0x0021, 0x0002);
    send_ctl(data, ugen, 0x03, 0x3c20, 0x003f, 0x0002);
    send_ctl(data, ugen, 0x03, 0x3d20, 0x0010, 0x0002);
    send_ctl(data, ugen, 0x03, 0x3e20, 0x0008, 0x0002);
    send_ctl(data, ugen, 0x03, 0x3f20, 0x000c, 0x0002);
    send_ctl(data, ugen, 0x03, 0x4020, 0x000c, 0x0002);
    send_ctl(data, ugen, 0x03, 0x4120, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x4220, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x4320, 0x004f, 0x0002);
    send_ctl(data, ugen, 0x03, 0x4420, 0x00ff, 0x0002);
    send_ctl(data, ugen, 0x03, 0x4620, 0x0020, 0x0002);
    send_ctl(data, ugen, 0x03, 0x4720, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x4820, 0x0090, 0x0002);
    send_ctl(data, ugen, 0x03, 0x4920, 0x00e6, 0x0002);
    send_ctl(data, ugen, 0x03, 0x4a20, 0x0002, 0x0002);
    send_ctl(data, ugen, 0x03, 0x4b20, 0x0054, 0x0002);
    send_ctl(data, ugen, 0x03, 0x4c20, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x5020, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x5120, 0x0058, 0x0002);
    send_ctl(data, ugen, 0x03, 0x5220, 0x0020, 0x0002);
    send_ctl(data, ugen, 0x03, 0x5420, 0x0057, 0x0002);
    send_ctl(data, ugen, 0x03, 0x5520, 0x00f1, 0x0002);
    send_ctl(data, ugen, 0x03, 0x5620, 0x0020, 0x0002);
    send_ctl(data, ugen, 0x03, 0x5720, 0x0070, 0x0002);
    send_ctl(data, ugen, 0x03, 0x5820, 0x0060, 0x0002);
    send_ctl(data, ugen, 0x03, 0x5c20, 0x0050, 0x0002);
    send_ctl(data, ugen, 0x03, 0x5d20, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x5f20, 0x0080, 0x0002);
    send_ctl(data, ugen, 0x03, 0x7020, 0x0018, 0x0002);
    send_ctl(data, ugen, 0x03, 0x7120, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x7220, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x7520, 0x0022, 0x0002);
    send_ctl(data, ugen, 0x03, 0x7620, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x7720, 0x0001, 0x0002);
    send_ctl(data, ugen, 0x03, 0x7c20, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x7d20, 0x0052, 0x0002);
    send_ctl(data, ugen, 0x03, 0xba20, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0xbb20, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0xbc20, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0xc220, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0xc720, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0xe420, 0x001a, 0x0002);
    send_ctl(data, ugen, 0x03, 0xe920, 0x0008, 0x0002);
    send_ctl(data, ugen, 0x03, 0xec20, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0xef20, 0x0000, 0x0002);
    /* */
    send_ctl(data, ugen, 0x02, 0x00a8, 0x0000, 0x0002);

    /* Random values */
    // send_ctl(data, ugen, 0x1a, 0x0000, 0x0000, 0x0010);

    /* ASIE5606 on */
    send_ctl(data, ugen, 0x08, 0x11ff, 0x0000, 0x0001);
    usleep (50 * 1000);
    send_ctl(data, ugen, 0x08, 0x1100, 0x0000, 0x0001);
    usleep (50 * 1000);
    send_ctl(data, ugen, 0x08, 0x11ff, 0x0000, 0x0001);
    usleep (100 * 1000);

    /* send_ctl(data, ugen, 0x02, 0x0400 | saddr, 0x0000, 0x0002); */
    send_ctl(data, ugen, 0x02, 0x0900 | saddr, 0x0000, 0x0002);
    if (data[0] != 1) {
      saddr = 0x4a;
      send_ctl(data, ugen, 0x02, 0x0900 | saddr, 0x0000, 0x0002);
      if (data[0] != 1) {
	// error
      }
    }
    if ( (data[1] & 7) == 6) {
      r5 = 0x80;
    }

    /* ASV5211 is high speed */
    send_ctl(data, ugen, 0x0a, 0x0000, 0x0000, 0x0001);
    usleep (30 * 1000);
    /* ASIE decrypt mode configuration */
    send_ctl(data, ugen, 0x03, 0x0500 | saddr, r5, 0x0002);

    if (log) *log << "done." << std::endl;
}

/**
 * 処理終了
 * 便宜上Endにしている。恐らくバッファクリア
 * @param fd 対象ファイルディスクリプタ
 */
void
Hdus::UsbProcEnd(int fd)
{
    if( ! saddr ) return;

    if (log) *log << "closing HDUS...";

    uint8_t data[0x200];
    int ugen = fd;

    /* Stop */
    /* stop ASV5211 bulk transfer */
    usleep(100 * 1000);
    send_ctl(data, ugen, 0x07, 0x0000, 0x0000, 0x0001);

    send_ctl(data, ugen, 0x09, 0x0100, 0x0000, 0x0001);
    send_ctl(data, ugen, 0x03, 0x8000 | saddr, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x0400 | saddr, 0x0000, 0x0002);
    send_ctl(data, ugen, 0x03, 0x0000 | saddr, 0x0000, 0x0002);

    /* Power off */
    usleep (50 * 1000);
    usleep (50 * 1000);
    send_ctl(data, ugen, 0x08, 0x0100, 0x0000, 0x0001);
    usleep (100 * 1000);
    send_ctl(data, ugen, 0x08, 0xdd0c, 0x0000, 0x0001);
    usleep (100 * 1000);

    saddr = 0;

    if (log) *log << "done." << std::endl;
}

/**
 * 初期化処理 : ストリーム制御データ
 * @param fd 対象ファイルディスクリプタ
 */
void
Hdus::UsbProcStreamInit(int fd)
{}

/**
 * ＬＥＤ制御
 * @param fd 対象ファイルディスクリプタ
 * @param color 色
 */
void
Hdus::UsbProcLED(int fd, BonLedColor color)
{}

/**
 * 初期化処理 : 固定処理B
 * @param fd 対象ファイルディスクリプタ
 */
void
Hdus::UsbProcFixInitB(int fd)
{}



void
Hdus::send_ctl(uint8_t *data, int fd, int req, int val, int idx, int len)
{
//    fprintf( stderr, "fd = 0x%x, req = 0x%x, val = 0x%x\n", fd, req, val );

    usbdevfs_ctrltransfer ctrl = {
        USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE, req, val, idx, len, REQUEST_TIMEOUT, data
    };
	
    usb_ctrl(fd, &ctrl);
}
